class Errors(list):
    """
    错误列表
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def error(self, message: str):
        self.append({'level': 'error', 'pos_stack': [], 'message': message})
        return self

    def warn(self, message: str):
        self.append({'level': 'warn', 'pos_stack': [], 'message': message})
        return self

    def pos(self, pos: str):
        for item in self:
            item['pos_stack'].append(pos)

    def __str__(self):
        ret = ''
        for item in self:
            ret += '[{}][{}] {}\n'.format(item["level"], ':'.join(reversed(item['pos_stack'])), item["message"])
        return ret


__all__ = (
    'Errors',
)



class AssAttr:
    _range: tuple = ...
    _attr_type: type = ...
    _default = None

    def __new__(cls, value=None):
        if value is None:
            return cls._default
        if cls._attr_type is not ...:
            value = cls._attr_type(value)
        if cls._attr_type == str:
            value = value.lower()
            for cmp_value in vars(cls).values():  # 忽略大小写比较字符串
                if isinstance(cmp_value, str) and cmp_value.lower() == value:
                    return cmp_value

        if value in vars(cls).values():
            return value
        if cls._range is not ...  \
           and (cls._range[0] is ... or cls._range[0] <= value)  \
           and (cls._range[1] is ... or cls._range[1] >= value):
            return value

        if cls._default is not None:
            return cls._default
        raise KeyError('Could not parse `{}` as {}'.format(value, cls.__name__))


__all__ = (
    'AssAttr',
)



import re


class AssColor:
    def __init__(self, value: int | str):
        """
        ASS 颜色类型

        参数：
            value: 颜色值
                - 当传入整数时，直接存储作为颜色值
                - 当传入字符串时，按照 AGBR 的顺序进行解析
        用法：
            dm_color = AssColor("&H66CCFF")  # 定义一个颜色，并将 66CCFF 作为值
            print(int(dm_color))  # 打印其整数值
            print(dm_color)  # 打印其十六进制值
            print(AssColor(17900) == AssColor("&H66CCFF"))  # 比较两个颜色值是否相同
        """
        if isinstance(value, int):
            self.__color_value = value
        elif isinstance(value, str):
            self.parse(value)
        else:
            raise Exception('Unsupported AssColor value type `{}`. '.format(type(value)))

    def parse(self, raw_str: str) -> None:
        """
        通过字符串改变当前颜色对象的颜色值

        参数：
            raw_str: 字符颜色值
        """
        try:
            raw_str = raw_str.strip('&Hh')
            self.__color_value = int(raw_str, base=16)
        except Exception as e:
            raise TypeError('Could not parse `{}` as ass color. {}'.
                            format(raw_str, e))

    def dump(self) -> str:
        """
        将颜色转为十六进制字符串

        返回值：
            字符串颜色值
        """
        alpha: int = (self.__color_value & 0xFF000000) >> 24
        color: int = (self.__color_value & 0x00FFFFFF)

        ret_str: str = '&H'
        if alpha != 0:  # 若透明度为 0 则不显示透明度的 0
            ret_str += '{:0>2X}'.format(alpha)
        ret_str += '{:0>6X}'.format(color)
        return ret_str

    def __eq__(self, other) -> bool:
        """
        判断两个颜色对象值是否相等

        返回值：
            颜色是否相同
        """
        if not isinstance(other, AssColor):
            raise Exception('Could not compare with `AssColor` and `{}`'.format(type(other)))
        return other.__color_value == self.__color_value

    def __str__(self) -> str:
        """
        将颜色转为十六进制字符串
        返回值：
            字符串颜色值
        """
        return self.dump()

    def __int__(self) -> int:
        """
        将颜色转为整形

        返回值：
            整数颜色值
        """
        return self.__color_value


class AssTime:
    _match_time = re.compile(r'(?P<h>\d+):(?P<m>\d+):(?P<s>\d+).(?P<ms>\d+)')

    def __init__(self, value):
        self.__time_value = 0
        if isinstance(value, str):
            self.parse(value)
        elif isinstance(value, AssTime):
            self.__time_value = float(value)
        else:
            self.__time_value = float(value)

    def parse(self, value: str):
        match = self._match_time.match(value)
        if match is None:
            raise Exception(f'Could not parse `{value}` as ass time. ')
        time_path = match.groupdict()
        self.__time_value =                 \
            int(time_path['h']) * 3600 +    \
            int(time_path['m']) * 60 +      \
            int(time_path['s']) +           \
            int(time_path['ms']) / 1000

    def dump(self):
        hour = int(self.__time_value // 3600)
        minute = int(self.__time_value // 60 % 60)
        second = int(self.__time_value % 60)
        milli_second = int(self.__time_value * 100 % 100)

        return f'{hour:0>2d}:{minute:0>2d}:{second:0>2d}.{milli_second:0>2d}'

    def __str__(self):
        return self.dump()

    def __float__(self):
        return self.__time_value


__all__ = (
    'AssColor',
    'AssTime',
)

t = AssTime(1.2592)
print(t)



import sys
import inspect
import re

from easyass.common.asstypes import AssColor


class TextBase:
    _prefix = ''  # 前缀
    _with_bracket = True  # 是否含有括号
    _arg_mapper: dict[str, dict] = ...  # 参数字典
    """
    _arg_mapper 字典定义函数参数
    key为参数名称，value为可迭代对象，其第一个元素表示 是否可选，第二个参数定义其 转换函数
    _arg_mapper = {
        'arg_name': (True, int),
        ...
    }
    """

    def __init__(self, *args, **kwargs):
        self._arg_values: dict = {arg_name: None for arg_name in self._arg_mapper} \
            if self._arg_mapper is not ... else {}  # 参数值存储

        if 'raw_str' in kwargs:  # 指定源字符串
            self.parse(kwargs.get('raw_str', ''))
        elif self._arg_mapper is not ...:  # 正常传参
            if len(args) > len(self._arg_mapper):
                raise TypeError('`{}()` takes at most {} arguments. '.
                                format(self.__class__.__name__, len(self._arg_mapper)))
            args = {name: value for name, value in zip(self._arg_mapper, args)}
            args.update(kwargs)
            self._handle_args(args)

    def _handle_args(self, args: dict[str, any]) -> None:
        for arg_name, arg_value in args.items():  # 填充已经填写的参数
            if arg_name not in self._arg_mapper:
                raise TypeError(arg_name)
            if arg_value is None:
                continue
            self._arg_values[arg_name] = \
                self._arg_mapper[arg_name][1](arg_value)

        if self._arg_mapper is ...:
            return
        for arg_name, arg_info in self._arg_mapper.items():  # 查找是否存在未赋值的必填参数
            if self._arg_values.get(arg_name, None) is None and not arg_info[0]:
                raise TypeError('In `{}()`, `{}` must be specified. '.
                                format(self.__class__.__name__, arg_name))

    def parse(self, ass_str: str):
        if self._arg_mapper is ...:
            return
        args_str = ass_str[len(self._prefix):].strip('() ')
        arg_values = args_str.split(',')
        args = {name: value for name, value in zip(self._arg_mapper, arg_values)}
        self._handle_args(args)

    def dump(self):
        arg_values = [str(value)
                      for value in self._arg_values.values()
                      if value is not None]
        args_str = ','.join(arg_values)
        if self._with_bracket:
            return '{}({})'.format(self._prefix, args_str)
        else:
            return '{}{}'.format(self._prefix, args_str)

    def __str__(self):
        return self.dump()

    def __getattr__(self, key):
        if key in self._arg_values:
            return self._arg_values[key]
        raise AttributeError(key)

    def __setattr__(self, key, value):
        if key in self.__dict__.get('_arg_values', {}):
            self.__dict__['_arg_values'][key] = value
            return
        super().__setattr__(key, value)

    def __add__(self, other):
        if not (isinstance(other, TextBase) or isinstance(other, str)):
            raise TypeError(f'unsupported operand types for +: '
                            f'`{self.__class__.__name__}` and `{type(other).__class__.__name__}`')
        return Text(self, other)

    def __radd__(self, other):
        return TextBase.__add__(other, self)


class Text(TextBase, list):
    __match_code_part = re.compile(r'{[^}]*?}')

    def __init__(self, *args):
        super().__init__()
        for arg in args:
            if isinstance(arg, Text):
                self.extend(arg)
            elif isinstance(arg, TextBase):
                self.append(arg)
            elif isinstance(arg, str):
                self.parse(arg)
            else:
                raise TypeError(arg)

    def parse(self, ass_str: str):
        matches = self.__match_code_part.finditer(ass_str)
        curr_pos = 0
        # 匹配 { } 花括号区域
        for match in matches:
            if match.start() != curr_pos:
                self.append(Str(raw_str=match.string[curr_pos: match.start()]))
            codes = match.string[match.start()+1: match.end()-1].split('\\')  # 切割各组代码
            for code in codes:
                if len(code) == 0:
                    continue
                for code_name, proc_type in _codes_mapper.items():
                    if code.startswith(code_name):
                        tar_type = proc_type
                        break
                else:
                    raise TypeError('Unknown code `{}`'.format(code))
                self.append(tar_type(raw_str=code))
            curr_pos = match.end()
        # 末端字符串
        if curr_pos != len(ass_str):
            self.append(Str(raw_str=ass_str[curr_pos:]))

    def dump(self) -> str:
        ret: str = ''
        for index, item in enumerate(self):
            if isinstance(item, Str):
                if index != 0 and not isinstance(self[index - 1], Str):  # 闭大括号
                    ret += '}'
            else:
                if index == 0 or isinstance(self[index - 1], Str):  # 开大括号
                    ret += '{'
                ret += '\\'
            ret += item.dump()
        return ret

    def __str__(self) -> str:
        return self.dump()


class Str(TextBase, str):
    """
    文本部分

    与 str 类无异
    """
    __escape_mapper = {
        r'\n': r'\{}n',
        r'\h': r'\{}h',
        ' ': r'\h',
        '\n': r'\N',
    }

    def __new__(cls, *args, **kwargs):
        if 'raw_str' in kwargs:  # 对 ass原始字符串进行反转义
            arg = cls.__unescape(kwargs.get('raw_str', ''))
            return super().__new__(cls, arg)
        else:
            return super().__new__(cls, *args, **kwargs)

    @staticmethod
    def __escape(raw_str: str) -> str:
        for from_str, to_str in Str.__escape_mapper.items():
            raw_str = raw_str.replace(from_str, to_str)
        return raw_str

    @staticmethod
    def __unescape(raw_str: str) -> str:
        for to_str, from_str in Str.__escape_mapper.items():
            raw_str = raw_str.replace(from_str, to_str)
        return raw_str

    def dump(self):
        return self.__escape(str.__str__(self))

    def __str__(self):
        return str.__str__(self)


class Bold(TextBase):
    """
    文本加粗

    参数:
     - enable: 0取消加粗 / 1加粗 / 其他表示字重
    """
    _prefix = 'b'
    _with_bracket = False
    _arg_mapper = {
        'enable': (False, int),
    }


class Italic(TextBase):
    """
    文本斜体

    参数:
     - enable: 0取消斜体 / 1启用斜体 / 其他表示字重
    """
    _prefix = 'i'
    _with_bracket = False
    _arg_mapper = {
        'enable': (False, int),
    }


class Underline(TextBase):
    """
    文本斜体

    参数:
     - enable: 0取消斜体 / 1启用斜体
    """
    _prefix = 'u'
    _with_bracket = False
    _arg_mapper = {
        'enable': (False, int),
    }


class Delete(TextBase):
    """
    文本删除线

    参数:
     - enable: 0取消删除线 / 1启用删除线
    """
    _prefix = 's'
    _with_bracket = False
    _arg_mapper = {
        'enable': (False, int),
    }


class Border(TextBase):
    """
    文本外边框

    参数:
     - width: 边框宽度
    """
    _prefix = 'bord'
    _with_bracket = False
    _arg_mapper = {
        'width': (False, float),
    }


class BorderX(Border):
    """
    文本外边框，仅 X 轴

    参数:
     - width: 边框宽度
    """
    _prefix = 'xbord'


class BorderY(Border):
    """
    文本外边框，仅 Y 轴

    参数:
     - width: 边框宽度
    """
    _prefix = 'ybord'


class Shadow(TextBase):
    """
    文本阴影

    参数:
     - depth: 阴影深度
    """
    _prefix = 'shad'
    _with_bracket = False
    _arg_mapper = {
        'depth': (False, float),
    }


class ShadowX(Shadow):
    """
    文本阴影，仅 X 轴

    参数:
     - depth: 阴影深度
    """
    _prefix = 'xshad'


class ShadowY(Shadow):
    """
    文本阴影，仅 Y 轴

    参数:
     - depth: 阴影深度
    """
    _prefix = 'xshad'


class FontName(TextBase):
    """
    文本字体名

    参数:
     - name: 字体名
    """
    _prefix = 'fn'
    _with_bracket = False
    _arg_mapper = {
        'name': (False, str),
    }


class FontSize(TextBase):
    """
    文本大小

    参数:
     - size: 大小，文本行高，单位 px
    """
    _prefix = 'fs'
    _with_bracket = False
    _arg_mapper = {
        'size': (False, int),
    }


class FontSizeInc(FontSize):
    """
    文本大小增大

    参数:
     - size: 增大倍数，文本行高会变成原来的 (1+size/10)，如 size=5 表示文本行高变为原来的 1.5 倍
    """
    _prefix = 'fs+'


class FontSizeDec(FontSize):
    """
    文本大小减小

    参数:
     - size: 减小倍数，文本行高会变成原来的 (1-size/10)，如 size=5 表示文本行高变为原来的 0.5 倍
    """
    _prefix = 'fs-'


class ScaleX(TextBase):
    """
    文本缩放，仅 X 轴

    参数:
     - ratio: 缩放比例
    """
    _prefix = 'fscx'
    _with_bracket = False
    _arg_mapper = {
        'ratio': (False, float),
    }


class ScaleY(TextBase):
    """
    文本缩放，仅 Y 轴

    参数:
     - ratio: 缩放比例
    """
    _prefix = 'fscy'
    _with_bracket = False
    _arg_mapper = {
        'ratio': (False, float),
    }


class Space(TextBase):
    """
    文字间距离

    参数:
     - size: 距离，单位 px
    """
    _prefix = 'fsp'
    _with_bracket = False
    _arg_mapper = {
        'size': (False, int),
    }


class Rotate(TextBase):
    """
    文本旋转，沿 Z 轴

    参数:
     - angle: 旋转角度
    """
    _prefix = 'fr'
    _with_bracket = False
    _arg_mapper = {
        'angle': (False, float),
    }


class RotateZ(Rotate):
    """
    文本旋转，沿 Z 轴

    参数:
     - angle: 旋转角度
    """
    _prefix = 'frz'


class RotateX(Rotate):
    """
    文本旋转，沿 X 轴

    参数:
     - angle: 旋转角度
    """
    _prefix = 'frx'


class RotateY(Rotate):
    """
    文本旋转，沿 Y 轴

    参数:
     - angle: 旋转角度
    """
    _prefix = 'fry'


class AngleX(TextBase):
    """
    文本倾斜，沿 X 轴

    参数:
     - factor: 倾斜因数，-0.5为斜体
    """
    _prefix = 'fax'
    _with_bracket = False
    _arg_mapper = {
        'factor': (False, float),
    }


class AngleY(TextBase):
    """
    文本倾斜，沿 Y 轴

    参数:
     - factor: 倾斜因数
    """
    _prefix = 'fay'
    _with_bracket = False
    _arg_mapper = {
        'factor': (False, float),
    }


class Encoding(TextBase):
    """
    文本字符集

    参数:
     - encoding: 字符集，0 英文 / 1 ANSI / 134 简体中文 / 136 繁体中文
    """
    _prefix = 'fe'
    _with_bracket = False
    _arg_mapper = {
        'encoding': (False, int),
    }


class Color(TextBase):
    """
    文本颜色

    参数:
     - color: 文本颜色值，可选，不填默认使用style中颜色
    """
    _prefix = 'c'
    _with_bracket = False
    _arg_mapper = {
        'color': (True, AssColor),
    }


class PrimaryColor(Color):
    """
    文本颜色，主颜色

    参数:
     - color: 文本颜色值，可选，不填默认使用style中颜色
    """
    _prefix = '1c'


class SecondaryColor(Color):
    """
    文本颜色，次颜色

    参数:
     - color: 文本颜色值，可选，不填默认使用style中颜色
    """
    _prefix = '2c'


class OutlineColor(Color):
    """
    文本颜色，边线颜色

    参数:
     - color: 文本颜色值，可选，不填默认使用style中颜色
    """
    _prefix = '3c'


class BackColor(Color):
    """
    文本颜色，背景颜色

    参数:
     - color: 文本颜色值，可选，不填默认使用style中颜色
    """
    _prefix = '4c'


class Alpha(TextBase):
    """
    文本透明度值，主透明度

    参数:
     - alpha: 文本透明度值
    """
    _prefix = 'alpha'
    _with_bracket = False
    _arg_mapper = {
        'alpha': (True, str),
    }


class PrimaryAlpha(Alpha):
    """
    文本透明度值，主透明度

    参数:
     - alpha: 文本透明度值
    """
    _prefix = '1a'


class SecondaryAlpha(Alpha):
    """
    文本透明度值，次透明度

    参数:
     - alpha: 文本透明度值
    """
    _prefix = '2a'


class OutlineAlpha(Alpha):
    """
    文本透明度值，描边透明度

    参数:
     - alpha: 文本透明度值
    """
    _prefix = '3a'


class BackAlpha(Alpha):
    """
    文本透明度值，背景透明度

    参数:
     - alpha: 文本透明度值
    """
    _prefix = '4a'


class Align(TextBase):
    """
    文本对齐

    参数:
     - alpha: 文本透明度值

    备注:
    与小键盘的布局一致
    7 8 9
    4 5 6
    1 2 3
    """
    _prefix = 'an'
    _with_bracket = False
    _arg_mapper = {
        'side': (False, int),
    }


class AlignEx(Align):
    """
    文本对齐

    参数:
     - alpha: 文本透明度值

    备注:
    与小键盘的布局一致
    9 10 11
    5 6 7
    1 2 3
    """
    _prefix = 'a'


class Wrap(TextBase):
    """
    文本换行方式

    参数:
     - style: 文本换行方式 0 自动但顶部长 / 1 行尾 / 2 不换行 / 3 自动但底部长
    """
    _prefix = 'q'
    _with_bracket = False
    _arg_mapper = {
        'style': (True, int),
    }


class ChangeStyle(TextBase):
    """
    文本换行方式

    参数:
     - name: 要更改的样式名，此代码后面的文本都会换成此样式
    """
    _prefix = 'r'
    _with_bracket = False
    _arg_mapper = {
        'name': (True, str),
    }


class Move(TextBase):
    """
    直线移动

    参数:
     - x1: 开始位置x，单位 px
     - x2: 开始位置y，单位 px
     - y1: 结束位置x，单位 px
     - y2: 结束位置y，单位 px
     - t1: 开始运动时间，单位毫秒，可选
     - t2: 结束运动时间，单位毫秒，可选
    """
    _prefix = 'move'
    _with_bracket = True
    _arg_mapper = {
        'x1': (False, int),
        'x2': (False, int),
        'y1': (False, int),
        'y2': (False, int),
        't1': (True, int),
        't2': (True, int),
    }


class Pos(TextBase):
    """
    固定停留在指定位置

    参数:
     - x: 位置x，单位 px
     - y: 位置y，单位 px
    """
    _prefix = 'pos'
    _with_bracket = True
    _arg_mapper = {
        'x': (False, int),
        'y': (False, int),
    }


class Org(TextBase):
    """
    指定定位点位置

    物体的旋转、移动等位置都是以定位点作为基准

    参数:
     - x: 位置x，单位 px
     - y: 位置y，单位 px
    """
    _prefix = 'org'
    _with_bracket = True
    _arg_mapper = {
        'x': (False, int),
        'y': (False, int),
    }


class Fade(TextBase):
    """
    淡入和淡出

    参数:
     - t1: 淡入时间，单位毫秒
     - t2: 淡出时间，单位毫秒
    """
    _prefix = 'fad'
    _with_bracket = True
    _arg_mapper = {
        't1': (False, int),
        't2': (False, int),
    }


class FadeEx(TextBase):
    """
    更加复杂的淡入和淡出效果

    参数:
     - a1: 时间 t1 之前的透明度
     - a2: 时间 t1 到 t2 透明度由 a1 渐变为 a2
     - a3: 时间 t3 到 t4 透明度由 a2 渐变为 a3
     - t1: 时间节点1，单位毫秒
     - t2: 时间节点2，单位毫秒
     - t3: 时间节点3，单位毫秒
     - t4: 时间节点4，单位毫秒
    """
    _prefix = 'fade'
    _with_bracket = True
    _arg_mapper = {
        'a1': (False, str),
        'a2': (False, str),
        'a3': (False, str),
        't1': (False, int),
        't2': (False, int),
        't3': (False, int),
        't4': (False, int),
    }


class Clip(TextBase):
    """
    淡入和淡出

    参数:
     - t1: 淡入时间，单位毫秒
     - t2: 淡出时间，单位毫秒
    """
    _prefix = 'clip'
    _with_bracket = True
    _arg_mapper = {
        't1': (False, str),
    }


class Draw(TextBase):
    """
    开启绘图模式，并设置绘图等级

    坐标系比例会变为原来的 1/(2^(level-1))，如果level为0表示退出绘图模式

    参数:
     - level: 绘图等级
    """
    _prefix = 'p'
    _with_bracket = False
    _arg_mapper = {
        't1': (False, str),
    }


class DrawBegin(TextBase):
    """
    开启绘图模式，相当于 Draw(1)

    后面的文本都会被解释为绘图指令

    参数:
     - level: 绘图等级
    """
    _prefix = 'p1'
    _with_bracket = False


class DrawEnd(TextBase):
    """
    关闭绘图模式，相当于 Draw(0)

    后面的文本都会被解释为普通文本

    参数:
     - level: 绘图等级
    """
    _prefix = 'p1'
    _with_bracket = False


# 注册全部 code
_codes = []
for _, cls in inspect.getmembers(sys.modules[__name__], inspect.isclass):
    if '_prefix' not in cls.__dict__ or not cls.__dict__.get('_prefix'):
        continue
    prefix = cls.__dict__.get('_prefix')
    _codes.append((prefix, cls))

# 将code按前缀长度排序
_codes.sort(key=lambda pair: len(pair[0]), reverse=True)
_codes_mapper: dict[str, type] = {_code[0]: _code[1] for _code in _codes}
del _codes



from .scriptinfo import *
from .events import *
from .styles import *
from .common.errors import Errors


class Ass:
    def __init__(self):
        self.script_info: ScriptInfo = ScriptInfo()
        self.styles: Styles = Styles()
        self.events: Events = Events()

        self.__parser_status: ScriptInfo | Styles | Events = self.script_info
        self.__parser_pos_name: str = ''

    def parse(self, ass_str: str) -> Errors:
        err: Errors = Errors()
        ass_str_lines = ass_str.split('\n')
        for line, ass_str_line in enumerate(ass_str_lines):
            line_err = self.parse_line(ass_str_line)
            line_err.pos(f'Line{line + 1}')
            err += line_err
        return err

    def parse_line(self, ass_str: str) -> Errors:
        err: Errors = Errors()
        ass_str = ass_str.lstrip()
        if len(ass_str) == 0:  # 空行
            return err
        if ass_str.startswith('['):
            ass_str_lower = ass_str.lower()
            if ass_str_lower.startswith(EVENTS_PART_TITLE.lower(), 1):
                self.__parser_status = self.events
                self.__parser_pos_name = 'event'
            elif ass_str_lower.startswith(STYLES_PART_TITLE.lower(), 1):
                self.__parser_status = self.styles
                self.__parser_pos_name = 'style'
            elif ass_str_lower.startswith(SCRIPT_INFO_PART_TITLE.lower(), 1):
                self.__parser_status = self.script_info
                self.__parser_pos_name = 'script_info'
        else:
            err += self.__parser_status.parse(ass_str)
            err.pos(self.__parser_pos_name)
        return err

    def dump(self):
        script_info_lines, script_info_errs = self.script_info.dump()
        styles_lines, styles_errs = self.styles.dump()
        events_lines, events_errs = self.events.dump()
        styles_errs.pos('style')
        events_errs.pos('event')
        lines = script_info_lines + [''] + styles_lines + [''] + events_lines
        script_info_errs += styles_errs + events_errs
        return lines, script_info_errs



from easyass.common.errors import Errors
from easyass.common.assattr import AssAttr
from easyass.common.asstypes import AssColor


class Styles(list):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.style_format: StyleFormat = StyleFormat()

    def parse(self, ass_str: str) -> Errors:
        err = Errors()
        ass_line = ass_str.split(':', 1)
        if len(ass_line) != 2:
            return err.error(f'Fail to parse `{ass_str}` as Styles. ')
        title = ass_line[0].strip()
        if title == FORMAT_LINE_TITLE:
            self.style_format = StyleFormat()
            format_err = self.style_format.parse(ass_str)
            format_err.pos('format')
        elif title == STYLE_LINE_TITLE:
            if self.style_format is ...:
                return err.error(f'`Format` line is missing. ')
            style_item = StyleItem()
            style_err = style_item.parse(ass_str, self.style_format)
            style_err.pos('item')
            err += style_err
            self.append(style_item)
        return err

    def dump(self) -> (list[str], Errors):
        errors = Errors()
        dump_lines: list[str] = ['[{}]'.format(STYLES_PART_TITLE)]

        format_lines, format_errors = self.style_format.dump()
        format_errors.pos('format')
        dump_lines += format_lines
        errors += format_errors
        for index, item in enumerate(self):
            item_line, item_error = item.dump(self.style_format)
            item_error.pos(f'item:{index}')
            dump_lines += item_line
            errors += item_error
        return dump_lines, errors


class StyleFormat(list):
    def __init__(self, *args):
        super().__init__()
        if not len(args):
            self.extend(STYLE_ATTR_DEF.keys())

    def parse(self, ass_str: str) -> Errors:
        err = Errors()
        self.clear()
        ass_line = ass_str.split(':', 1)
        if len(ass_line) != 2 or ass_line[0].strip() != FORMAT_LINE_TITLE:
            return err.error(f'Fail to parse `{ass_str}` as style format. ')
        for style_attr in ass_line[1].split(','):
            style_attr = style_attr.strip()
            if style_attr not in STYLE_ATTR_DEF:
                return err.error(f'Unknown style attribute `{style_attr}`. ')
            self.append(style_attr)
        return err

    def dump(self) -> (list[str], Errors):
        err = Errors()
        if not len(self):
            return [], err.error('Missing `style_attrs`')
        format_line = FORMAT_LINE_TITLE + ':' + ','.join(self)
        return [format_line], Errors()


class StyleItem:
    def __init__(self, **kwargs):
        self.style_attrs = {attr_name: None for attr_name in STYLE_ATTR_DEF.keys()}
        for key, value in kwargs.items():  # 支持构造时传初始值
            if key in self.style_attrs:
                self.style_attrs[key] = value

    def parse(self, ass_str: str, style_format: StyleFormat) -> Errors:
        err = Errors()
        ass_line = ass_str.split(':', 1)
        if len(ass_line) != 2 or ass_line[0].strip() != STYLE_LINE_TITLE:
            return err.error(f'Fail to parse `{ass_str}` as style values. ')
        style_values = ass_line[1].split(',')
        if len(style_values) != len(style_format):
            return err.error(f'`Styles` line does not match `Format` line. ')
        for index, style_attr in enumerate(style_format):
            try:
                self.style_attrs[style_attr] = \
                    STYLE_ATTR_DEF[style_attr](style_values[index])
            except Exception as exception:
                err.error(f'Could not parse `{style_values[index]}` '
                          f'as `{style_attr}`. Exception: {exception}. ')
        return err

    def dump(self, style_format: StyleFormat) -> (list[str], Errors):
        err = Errors()
        style_values = []
        for style_attr in style_format:
            if self.style_attrs[style_attr] is None:
                err.error(f'Style `{style_attr}` not specified. ')
            style_values.append(str(self.style_attrs[style_attr]))

        styles_line = STYLE_LINE_TITLE + ':' + ','.join(style_values)
        return [styles_line], err

    def __getattr__(self, attribute):
        if attribute in self.style_attrs:
            return self.style_attrs[attribute]
        raise AttributeError(attribute)

    def __setattr__(self, key, value):
        style_attrs = self.__dict__.get('style_attrs', {})
        if key in style_attrs:
            style_attrs[key] = STYLE_ATTR_DEF[key](value)
        else:
            super().__setattr__(key, value)


class Bold(AssAttr):
    _attr_type = int
    enable = -1
    disable = 0
    _default = -1


class Italic(AssAttr):
    _attr_type = int
    enable = -1
    disable = 0
    _default = -1


class Underline(AssAttr):
    _attr_type = int
    enable = -1
    disable = 0
    _default = -1


class StrikeOut(AssAttr):
    _attr_type = int
    enable = -1
    disable = 0
    _default = -1


class BorderStyle(AssAttr):
    _attr_type = int
    outlineAndShadow = 1
    opaque = 3
    _default = 1


class Alignment(AssAttr):
    _attr_type = int
    NW = 7
    N = 8
    NE = 9
    W = 4
    CENTER = 5
    E = 6
    SW = 1
    S = 2
    SE = 3
    _default = 5


STYLES_PART_TITLE = 'v4+ Styles'
FORMAT_LINE_TITLE = 'Format'
STYLE_LINE_TITLE = 'Style'
STYLE_ATTR_DEF = {
    'Name': str,  # 样式名称
    'Fontname': str,  # 字体名称
    'Fontsize': int,  # 文字大小
    'PrimaryColour': AssColor,  # 主颜色
    'SecondaryColour': AssColor,  # 次颜色
    'OutlineColour': AssColor,  # 边框颜色
    'BackColour': AssColor,  # 边框颜色
    'Bold': Bold,  # 粗体
    'Italic': Italic,  # 斜体
    'Underline': Underline,  # 下划线
    'StrikeOut': StrikeOut,  # 删除线
    'ScaleX': float,  # 文本宽度
    'ScaleY': float,  # 文本高度
    'Spacing': float,  # 字符之间的间隙
    'Angle': float,  # 文本旋转角度
    'BorderStyle': BorderStyle,  # 边框样式
    'Outline': float,  # 边框宽度
    'Shadow': int,  # 阴影宽度
    'Alignment': Alignment,  # 对齐方式 与小键盘一致
    'MarginL': int,  # 左边距
    'MarginR': int,  # 右边距
    'MarginV': int,  # 垂直边距
    'Encoding': int,  # 编码方式
}

__all__ = (
    'Styles',
    'StyleFormat',
    'StyleItem',
    'Bold',
    'Italic',
    'Underline',
    'StrikeOut',
    'BorderStyle',
    'Alignment',
    'STYLES_PART_TITLE',
    'FORMAT_LINE_TITLE',
    'STYLE_LINE_TITLE',
    'STYLE_ATTR_DEF'
)



from easyass.common.errors import Errors
from easyass.common.assattr import AssAttr


class ScriptInfo:
    def __init__(self):
        self.script_info_attrs = {attr_name: None
                                  for attr_name in SCRIPT_INFO_ATTR_DEF.keys()}
        self.script_info_attrs.update({
            'ScriptType': 'V4.00+',
            'Title': '<untitled>'
        })

    def parse(self, ass_str: str) -> Errors:
        err = Errors()
        str_part = ass_str.split(':')
        if len(str_part) != 2:
            return err.error(f'Fail to parse `{ass_str}` as script info. ')
        attribute = str_part[0].strip()
        value = str_part[1].strip()
        if attribute not in self.script_info_attrs:
            return err.error(f'Unknown script info attribute `{attribute}`. ')
        try:
            attribute_type = SCRIPT_INFO_ATTR_DEF[attribute]
            self.script_info_attrs[attribute] = attribute_type(value)
        except Exception as exception:
            return err.error(f'Could not parse `{value}` as `{attribute}`. '
                             f'Exception: {exception}. ')
        return err

    def dump(self) -> (list[str], Errors):
        errors: Errors = Errors()
        dump_lines: list[str] = [
            '[{}]'.format(SCRIPT_INFO_PART_TITLE)
        ]
        for key, value in self.script_info_attrs.items():
            if value is not None:
                dump_lines.append('{}: {}'.format(key, value))
        return dump_lines, errors

    def __str__(self):
        return '\n'.join(self.dump()[0])

    def __getattr__(self, key):
        if key in self.script_info_attrs:
            return self.script_info_attrs[key]
        raise AttributeError(key)

    def __setattr__(self, key, value):
        script_info_attrs = self.__dict__.get('script_info_attrs', {})
        if key in script_info_attrs:
            script_info_attrs[key] = SCRIPT_INFO_ATTR_DEF[key](value)
        else:
            super().__setattr__(key, value)


class Collisions(AssAttr):
    _attr_type = str
    normal: str = 'Normal'
    reverse: str = 'Reverse'


class WrapStyle(AssAttr):
    _attr_type = int
    auto_but_top_line_longer: int = 0
    last_word_of_line: int = 1
    never_swap: int = 2
    auto_but_bottom_line_longer: int = 3


class ScaledBorderAndShadow(AssAttr):
    _attr_type = str
    yes: str = 'Yes'
    no: str = 'No'


SCRIPT_INFO_PART_TITLE = 'Script Info'
SCRIPT_INFO_ATTR_DEF = {
    'ScriptType': str,  # 脚本版本
    'Title': str,  # 标题
    'PlayResX': float,  # 分辨率 X
    'PlayResY': float,  # 分辨率 Y
    'Collisions': Collisions,  # 重叠处理
    'PlayDepth': int,  # 颜色深度
    'Timer': float,  # 计时器速度
    'WrapStyle': WrapStyle,  # 换行方式
    'ScaledBorderAndShadow': ScaledBorderAndShadow,  # 边框与阴影是否随分辨率缩放
    'OriginalScript': str,  # 原始作者
    'OriginalTranslation': str,  # 原始翻译者 可选
    'OriginalEditing': str,  # 原始编辑者 可选
    'OriginalTiming': str,  # 原始打轴者 可选
    'SynchPoint': float,   # 开始播放时间 可选
    'ScriptUpdatedBy': str,  # 脚本更新者 可选
    'UpdateDetails': str,  # 更新细节 可选
}


__all__ = (
    'ScriptInfo',
    'Collisions',
    'WrapStyle',
    'ScaledBorderAndShadow',
    'SCRIPT_INFO_PART_TITLE',
)



from enum import Enum
from typing import SupportsIndex

from easyass.events.text import Text

from easyass.common.asstypes import AssTime
from easyass.common.errors import Errors


class EventTypes(Enum):
    dialogue = 'Dialogue'
    comment = 'Comment'
    picture = 'Picture'
    sound = 'Sound'
    movie = 'Movie'
    command = 'Command'


class Events(list):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.event_format: EventFormat = EventFormat()

    def parse(self, ass_str: str) -> Errors:
        """
        解析事件字符串生成事件数组

        参数：
            ass_str: 需要解析的事件字符串，含有多条事件
        返回值：
            解析的时候发生的错误
        """
        err = Errors()
        ass_line = ass_str.split(':', 1)
        if len(ass_line) != 2:
            return err.error(f'Fail to parse `{ass_str}` as Styles. ')
        title = ass_line[0].strip()
        if title == FORMAT_LINE_TITLE:
            self.event_format = EventFormat()
            format_err = self.event_format.parse(ass_str)
            format_err.pos('format')
            err += format_err
        elif title == str(EventTypes.dialogue.value):
            if self.event_format is ...:
                return err.error('`Format` line is missing. ')
            event_item = EventItem()
            style_err = event_item.parse(ass_str, self.event_format)
            style_err.pos('item')
            err += style_err
            self.append(event_item)
        return err

    def dump(self) -> (list[str], Errors):
        """
        将当前事件数组格式化为ass文本

        返回值：
            ass 文本数组
            格式化时发生的错误
        """
        errors: Errors = Errors()
        dump_lines: list[str] = [
            '[{}]'.format(EVENTS_PART_TITLE)
        ]
        format_line, format_err = self.event_format.dump()  # format 行
        format_err.pos('format')
        dump_lines += format_line
        errors += format_err

        for index, item in enumerate(self):  # item 行
            item_line, item_error = item.dump(self.event_format)
            item_error.pos(f'item:{index}')
            dump_lines += item_line
            errors += item_error
        return dump_lines, errors

    def append(self, *args, **kwargs) -> None:
        """
        向事件列表末尾追加一个事件对象 EventItem

        1. 当参数只有一个时，参数类型必须为 EventItem 对象；
        2. 有多个参数时，将根据参数构造 EventItem 对象，具体参数如下：
            参数：
                event_type: EventTypes 事件类型，枚举
            关键字参数：
                Marked: int 是否已标识
                Layer: int 图层 大的置于顶层
                Start: AssTime 开始时间
                End: AssTime 结束时间
                Style: str 使用的样式名称 可选
                Name: str 角色名 可选
                MarginL: int 左边距覆写值 px 可选
                MarginR: int 右边距覆写值 px 可选
                MarginV: int 垂直边距覆写值 px 可选
                Effect: str 过渡效果 (暂不做特殊支持) 可选
                Text: Text 文本
        """
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], EventItem):
            super().append(args[0])
        else:
            super().append(EventItem(*args, **kwargs))

    def insert(self, index: SupportsIndex, *args, **kwargs) -> None:
        """
        向事件列表中插入一个事件对象 EventItem

        1. 当参数只有两个时，具体参数如下；
            参数：
                index: 插入位置
                event_item: 需要插入的 EventItem 对象
        2. 有三个及以上个参数时，将根据参数构造 EventItem 对象，具体参数如下：
            参数：
                index: 插入位置
                event_type: EventTypes 事件类型，枚举
            关键字参数：
                Marked: int 是否已标识
                Layer: int 图层 大的置于顶层
                Start: AssTime 开始时间
                End: AssTime 结束时间
                Style: str 使用的样式名称 可选
                Name: str 角色名 可选
                MarginL: int 左边距覆写值 px 可选
                MarginR: int 右边距覆写值 px 可选
                MarginV: int 垂直边距覆写值 px 可选
                Effect: str 过渡效果 (暂不做特殊支持) 可选
                Text: Text 文本
        """
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], EventItem):
            super().insert(index, args[0])
        else:
            super().insert(index, EventItem(*args, **kwargs))


class EventFormat(list):
    def __init__(self, *args):
        super().__init__(*args)
        if not len(args):
            self.extend(EVENT_ATTR_DEF.keys())

    def parse(self, ass_str: str) -> Errors:
        err = Errors()
        self.clear()
        ass_line = ass_str.split(':', 1)
        if len(ass_line) != 2 or ass_line[0].strip() != FORMAT_LINE_TITLE:
            return err.error(f'Fail to parse `{ass_str}` as event format. ')
        style_attrs = ass_line[1].split(',')
        for event_attr in style_attrs:
            event_attr = event_attr.strip()
            if event_attr not in EVENT_ATTR_DEF:
                return err.error(f'Unknown event attribute `{event_attr}`. ')
            self.append(event_attr)
        return err

    def dump(self) -> (list[str], Errors):
        format_line = FORMAT_LINE_TITLE + ':' + ','.join(self)
        return [format_line], Errors()


class EventItem:
    def __init__(self, event_type: EventTypes = ..., **kwargs):
        """
        ass事件项
        可以通过参数指定初始值

        参数：
            event_type: EventTypes 事件类型，枚举
        关键字参数：
            Marked: int 是否已标识
            Layer: int 图层 大的置于顶层
            Start: AssTime 开始时间
            End: AssTime 结束时间
            Style: str 使用的样式名称 可选
            Name: str 角色名 可选
            MarginL: int 左边距覆写值 px 可选
            MarginR: int 右边距覆写值 px 可选
            MarginV: int 垂直边距覆写值 px 可选
            Effect: str 过渡效果 (暂不做特殊支持) 可选
            Text: Text 文本
        """
        self.event_attrs = {attr_name: attr_info[1] for attr_name, attr_info in EVENT_ATTR_DEF.items()}
        self.event_type: EventTypes = event_type
        for key, value in kwargs.items():  # 支持构造时传初始值
            if key in self.event_attrs:
                self.event_attrs[key] = EVENT_ATTR_DEF[key][0](value)

    def parse(self, ass_str: str, event_format: EventFormat) -> Errors:
        err = Errors()
        ass_line = ass_str.split(':', 1)
        if len(ass_line) != 2:
            return err.error(f'Fail to parse `{ass_str}` as style values. ')
        # 解析类型
        event_type = ass_line[0].strip()
        if EventTypes(event_type) not in EventTypes:
            return err.error(f'Unknown event type `{event_type}`. ')
        self.event_type = EventTypes(event_type)

        # 解析属性
        event_values = ass_line[1].split(',', len(event_format) - 1)
        if len(event_values) < len(event_format):
            return err.error(f'`Styles` line does not match `Format` line. ')
        for index, style_attr in enumerate(event_format):
            try:
                self.event_attrs[style_attr] = \
                    EVENT_ATTR_DEF[style_attr][0](event_values[index])
            except Exception as exception:
                return err.error(f'Could not parse `{event_values[index]}` '
                                 f'as `{style_attr}`. Exception: {exception}. ')
        return err

    def dump(self, style_format: EventFormat) -> (list[str], Errors):
        err = Errors()
        style_values = []
        if self.event_type is ...:
            return [], err.error('`event_type` must be specified')
        for style_attr in style_format:
            if self.event_attrs[style_attr] is None:
                err.error(f'Style `{style_attr}` not specified. ')
            else:
                style_values.append(str(self.event_attrs[style_attr]))

        events_line = self.event_type.value + ':' + ','.join(style_values)
        return [events_line], err

    def __getattr__(self, attribute):
        if attribute in self.event_attrs:
            return self.event_attrs[attribute]
        raise AttributeError(attribute)

    def __setattr__(self, key, value):
        if key in self.__dict__.get('event_attrs', {}):
            self.__dict__['event_attrs'][key] = EVENT_ATTR_DEF[key][0](value)
        else:
            super().__setattr__(key, value)


EVENTS_PART_TITLE = 'Events'
FORMAT_LINE_TITLE = 'Format'
EVENT_ATTR_DEF = {
    'Marked': (int, 0),  # 是否已标识
    'Layer': (int, 0),  # 图层 大的置于顶层
    'Start': (AssTime, None),  # 开始时间
    'End': (AssTime, None),  # 结束时间
    'Style': (str, 'default'),  # 使用的样式名称
    'Name': (str, ''),  # 角色名
    'MarginL': (int, 0),  # 左边距覆写值 px
    'MarginR': (int, 0),  # 右边距覆写值 px
    'MarginV': (int, 0),  # 垂直边距覆写值 px
    'Effect': (str, ''),  # 过渡效果 (暂不做特殊支持
    'Text': (Text, Text()),  # 文本
}


__all__ = (
    'Events',
    'EventFormat',
    'EventItem',
    'EventTypes',
    'EVENTS_PART_TITLE',
    'FORMAT_LINE_TITLE',
    'EVENT_ATTR_DEF',
)



from .easyass import Ass

from .common.asstypes import AssColor, AssTime
from .common.errors import Errors

from .events.events import *
from .events import text
from .events.text import *

from .scriptinfo.scriptinfo import *

from .styles.styles import *



from setuptools import setup, find_packages
from easyass.__version__ import __version_str__

with open(r'README.md', 'r', encoding='utf-8') as fp:
    long_description = fp.read()

setup(
    name='easyass',
    version=__version_str__,
    keywords=['ass', 'subtitle'],
    description='An ass subtitle parsing library',
    license='MIT Licence',
    long_description=long_description,
    long_description_content_type="text/markdown",

    url='https://github.com/hihkm/easyAss',
    author='tikm',
    author_email='hkm@tikm.org',

    include_package_data=True,
    platforms='all',
    install_requires=[],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    packages=find_packages(),
    python_requires='>=3.7',
)
